@@ -7,13 +7,11 @@ import rtgym
 # local imports
 import tmrl.config.config_constants as cfg
 from tmrl.training_offline import TrainingOffline
-from tmrl.custom.custom_gym_interfaces import TM2020Interface, TM2020InterfaceLidar, TMInterface, TMInterfaceLidar
-from tmrl.custom.custom_memories import MemoryTM2020RAM, MemoryTMNF, MemoryTMNFLidar, TrajMemoryTMNFLidar, get_local_buffer_sample, get_local_buffer_sample_tm20_imgs
-from tmrl.custom.custom_preprocessors import obs_preprocessor_tm_act_in_obs, obs_preprocessor_tm_lidar_act_in_obs
+from tmrl.custom.custom_gym_interfaces import TM2020Interface, TM2020InterfaceLidar, TM2020InterfaceLidarProgress, TMInterface, TMInterfaceLidar
+from tmrl.custom.custom_memories import MemoryTM2020, MemoryTMNFLidar, MemoryTMNFLidarProgress, get_local_buffer_sample_lidar, get_local_buffer_sample_lidar_progress, get_local_buffer_sample_tm20_imgs
+from tmrl.custom.custom_preprocessors import obs_preprocessor_tm_act_in_obs, obs_preprocessor_tm_lidar_act_in_obs,obs_preprocessor_tm_lidar_progress_act_in_obs
 from tmrl.envs import GenericGymEnv
-# from tmrl.custom.custom_models import Mlp, MlpPolicy
-from tmrl.custom.custom_models import MLPActorCritic, REDQMLPActorCritic, RNNActorCritic, SquashedGaussianMLPActor, SquashedGaussianRNNActor
-# from tmrl.sac import SacAgent as SAC_Agent
+from tmrl.custom.custom_models import SquashedGaussianMLPActor, MLPActorCritic, REDQMLPActorCritic, RNNActorCritic, SquashedGaussianRNNActor, SquashedGaussianCNNActor, CNNActorCritic
 from tmrl.custom.custom_algorithms import SpinupSacAgent as SAC_Agent
 from tmrl.custom.custom_algorithms import REDQSACAgent as REDQ_Agent
 from tmrl.util import partial
@@ -26,17 +24,28 @@ assert ALG_NAME in ["SAC", "REDQSAC"], f"If you wish to implement {ALG_NAME}, do
 
 # MODEL, GYM ENVIRONMENT, REPLAY MEMORY AND TRAINING: ===========
 
-assert cfg.PRAGMA_LIDAR
-if cfg.PRAGMA_RNN:
-    assert ALG_NAME == "SAC", f"{ALG_NAME} is not implemented here."
-    TRAIN_MODEL = RNNActorCritic
-    POLICY = SquashedGaussianRNNActor
+if cfg.PRAGMA_LIDAR:
+    if cfg.PRAGMA_RNN:
+        assert ALG_NAME == "SAC", f"{ALG_NAME} is not implemented here."
+        TRAIN_MODEL = RNNActorCritic
+        POLICY = partial(SquashedGaussianRNNActor, act_buf_len=cfg.ACT_BUF_LEN)
+    else:
+        TRAIN_MODEL = MLPActorCritic if ALG_NAME == "SAC" else REDQMLPActorCritic
+        POLICY = partial(SquashedGaussianMLPActor, act_buf_len=cfg.ACT_BUF_LEN)
 else:
-    TRAIN_MODEL = MLPActorCritic if ALG_NAME == "SAC" else REDQMLPActorCritic
-    POLICY = SquashedGaussianMLPActor
+    assert not cfg.PRAGMA_RNN, "RNNs not supported yet"
+    assert ALG_NAME == "SAC", f"{ALG_NAME} is not implemented here."
+    TRAIN_MODEL = CNNActorCritic
+    POLICY = SquashedGaussianCNNActor
 
 if cfg.PRAGMA_LIDAR:
-    INT = partial(TM2020InterfaceLidar, img_hist_len=cfg.IMG_HIST_LEN, gamepad=cfg.PRAGMA_GAMEPAD) if cfg.PRAGMA_TM2020_TMNF else partial(TMInterfaceLidar, img_hist_len=cfg.IMG_HIST_LEN)
+    if cfg.PRAGMA_TM2020_TMNF:
+        if cfg.PRAGMA_PROGRESS:
+            INT = partial(TM2020InterfaceLidarProgress, img_hist_len=cfg.IMG_HIST_LEN, gamepad=cfg.PRAGMA_GAMEPAD)
+        else:
+            INT = partial(TM2020InterfaceLidar, img_hist_len=cfg.IMG_HIST_LEN, gamepad=cfg.PRAGMA_GAMEPAD)
+    else:
+        INT = partial(TMInterfaceLidar, img_hist_len=cfg.IMG_HIST_LEN)
 else:
     INT = partial(TM2020Interface, img_hist_len=cfg.IMG_HIST_LEN, gamepad=cfg.PRAGMA_GAMEPAD) if cfg.PRAGMA_TM2020_TMNF else partial(TMInterface, img_hist_len=cfg.IMG_HIST_LEN)
 
@@ -47,9 +56,21 @@ for k, v in CONFIG_DICT_MODIFIERS.items():
     CONFIG_DICT[k] = v
 
 # to compress a sample before sending it over the local network/Internet:
-SAMPLE_COMPRESSOR = get_local_buffer_sample if cfg.PRAGMA_LIDAR else get_local_buffer_sample_tm20_imgs
+if cfg.PRAGMA_LIDAR:
+    if cfg.PRAGMA_PROGRESS:
+        SAMPLE_COMPRESSOR = get_local_buffer_sample_lidar_progress
+    else:
+        SAMPLE_COMPRESSOR = get_local_buffer_sample_lidar
+else:
+    SAMPLE_COMPRESSOR = get_local_buffer_sample_tm20_imgs
 # to preprocess observations that come out of the gym environment and of the replay buffer:
-OBS_PREPROCESSOR = obs_preprocessor_tm_lidar_act_in_obs if cfg.PRAGMA_LIDAR else obs_preprocessor_tm_act_in_obs
+if cfg.PRAGMA_LIDAR:
+    if cfg.PRAGMA_PROGRESS:
+        OBS_PREPROCESSOR = obs_preprocessor_tm_lidar_progress_act_in_obs
+    else:
+        OBS_PREPROCESSOR = obs_preprocessor_tm_lidar_act_in_obs
+else:
+    OBS_PREPROCESSOR = obs_preprocessor_tm_act_in_obs
 # to augment data that comes out of the replay buffer (applied after observation preprocessing):
 SAMPLE_PREPROCESSOR = None
 
@@ -59,14 +80,17 @@ if cfg.PRAGMA_LIDAR:
     if cfg.PRAGMA_RNN:
         assert False, "not implemented"
     else:
-        MEM = MemoryTMNFLidar
+        if cfg.PRAGMA_PROGRESS:
+            MEM = MemoryTMNFLidarProgress
+        else:
+            MEM = MemoryTMNFLidar
 else:
-    MEM = MemoryTM2020RAM if cfg.PRAGMA_TM2020_TMNF else MemoryTMNF
+    assert cfg.PRAGMA_TM2020_TMNF, "TMNF is not officially supported."
+    MEM = MemoryTM2020
 
 MEMORY = partial(MEM,
                  memory_size=cfg.TMRL_CONFIG["MEMORY_SIZE"],
                  batch_size=cfg.TMRL_CONFIG["BATCH_SIZE"],
-                 obs_preprocessor=OBS_PREPROCESSOR,
                  sample_preprocessor=SAMPLE_PREPROCESSOR,
                  dataset_path=cfg.DATASET_PATH,
                  imgs_obs=cfg.IMG_HIST_LEN,
@@ -142,14 +166,16 @@ else:  # images
         TrainingOffline,
         env_cls=ENV_CLS,
         memory_cls=MEMORY,
-        epochs=100000,  # 10
-        rounds=10,  # 50
-        steps=50,  # 2000
-        update_model_interval=50,
-        update_buffer_interval=1,
-        max_training_steps_per_env_step=1.0,
+        epochs=cfg.TMRL_CONFIG["MAX_EPOCHS"],
+        rounds=cfg.TMRL_CONFIG["ROUNDS_PER_EPOCH"],
+        steps=cfg.TMRL_CONFIG["TRAINING_STEPS_PER_ROUND"],
+        update_model_interval=cfg.TMRL_CONFIG["UPDATE_MODEL_INTERVAL"],
+        update_buffer_interval=cfg.TMRL_CONFIG["UPDATE_BUFFER_INTERVAL"],
+        max_training_steps_per_env_step=cfg.TMRL_CONFIG["MAX_TRAINING_STEPS_PER_ENVIRONMENT_STEP"],
         profiling=cfg.PROFILE_TRAINER,
-        training_agent_cls=AGENT)
+        training_agent_cls=AGENT,
+        agent_scheduler=None,  # sac_v2_entropy_scheduler
+        start_training=cfg.TMRL_CONFIG["ENVIRONMENT_STEPS_BEFORE_TRAINING"])
 
 # CHECKPOINTS: ===================================================
 
